##    
# @file Vein_Biometric_Recognition_on_a_Smartphone-Python_Code.py
# @author Raul Garcia Martin and Raul Sanchez-Reillo
# @brief Program to generate two '.txt' files with the comparison scores (respresented by a number of coincident points)
#        between the one user pattern and its corresponding samples (Mated comparison)
#        and between the user pattern and its not corresponding samples (Non-Mated comparison).
#        The SIFT algorithm is used. The images should be in grayscale processing status. One user is define for
#        one pattern, right or left, regardless if it onws to the same person.
# @date 08/05/2020
# @version 1.0
#          \n (3.5.4 Python Version, 3.4.0 OpenCV Version on win64). The non-free version of OpenCV is needed
#          \n (pip install opencv-python-nonfree).


# LIBRARIES

import numpy as np
import cv2
import os
import csv
from threading import Thread, Lock, Timer
import time

# VARIABLES

## Images right pattern extension.
pattern_extension_right = "R.jpg"
## Images right left extension.
pattern_extension_left = "L.jpg"
## Lowe's ratio used for matching (0.8 recomended).
Lowes_ratio_test = 0.8
## Mated txt file name.
mated_txt_name = "Vein_Biometric_Recognition_on_a_Smartphone-SIFT_Mated.txt"
## Non-Mated txt file name.
non_mated_txt_name = "Vein_Biometric_Recognition_on_a_Smartphone-SIFT_Non-Mated.txt"
## Number of SIFT keypoints obtained for one image.
number_of_keypoints = 100

# FUNCTIONS

## Funtion to obtain the desired contrast of the image using the CLAHE algorithm several times. 
#  @arg img Image object to processed.
#  @arg num Number of interations.
#  @return The processed image object.
def multi_clahe(img, num):
    for i in range(num):
        img = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(4+i*2,4+i*2)).apply(img)
    return img

# CLASSES

## Class to define all users with its name (name of the photos files) and patterns (left and right wrist photos in the program folder,.jpg). 
#
class DefineUsers:
    ## Constructor
    #  @arg self The object pointer.
    #  @return Void.
    def __init__(self):
        ## @var name.
        # User's name.
        self.name = "User"
        
        ## @var number_of_users.
        # Number of users registered.
        self.number_of_users = 0  
        
        ## @var users.
        # Array of user object.
        self.users = []
        
        ## @var pattern.
        # Pattern number.
        self.pattern = ""
        
    ## Function to define the user objects.    
    #  @arg self The object pointer.
    #  @return Void.    
    def define_users(self):
        
        # Search users files (patterns) and users definitions.
        for i in os.listdir('.'):
            if i.endswith(pattern_extension_left) or i.endswith(pattern_extension_right):
                self.name = i[0 : 4]
                self.pattern = i
                self.number_of_users = self.number_of_users + 1
                self.users.append(User(self.name, self.pattern))
                
        # Search users samples and assignament to their users. Image contrast increse with CLAHE (2) and filtering with Gaussian, Median and Blur filters.
        for x in range (self.number_of_users):
            for i in os.listdir('.'):
                if (i.startswith(self.users[x].name)) and (not i.endswith(pattern_extension_left)) and (not i.endswith(pattern_extension_right)):
                    img = multi_clahe(cv2.imread(i,0), 2)
                    img = cv2.GaussianBlur(img,(11,11),0)
                    img = cv2.medianBlur(img,11)
                    img = cv2.blur(img,(11,11))
                    
                    self.users[x].samples.append(img)
                    self.users[x].sample_name.append(i[i.find("S") : i.find(".jpg")])
        
        # Run keypoints timer to know the time to obtain the keypoints of the all samples. 
        start_kp = time.time()
        
        # Define keypoints.
        for x in range (self.number_of_users):
            self.users[x].define_key_points()
        
        # Stop keypoints timer to know the time to obtain the keypoints of the all samples.
        end_kp = time.time()
        print("KP TIME(s):", end_kp - start_kp)
        
    ## Function to return the object with all the users.
    #  @arg self The object pointer.
    #  @return The object pointer.
    def get_users(self):
        return self

## User class with its name (name of the photos files) and patterns (left and right wrist photos in the program folder,.jpg).  
#                
class User:
    ## Constructor (name, pattern (.jpg), SIFT features with its number of keypoints and the keypoints and descriptors for the patterns).
    #  @arg self The object pointer.
    #  @arg name User's name.
    #  @arg name User's pattern.
    #  @return Void.
    def __init__(self, name, pattern):
        ## @var name.
        # User's name.
        self.name = name
        
        ## @var pattern.
        # User's processed image pattern. Image contrast increse with CLAHE (2) and filtering with Gaussian, Median and Blur filters.
        self.pattern = cv2.imread(pattern,0)
        self.pattern = multi_clahe(self.pattern, 2)
        self.pattern = cv2.GaussianBlur(self.pattern,(11,11),0)
        self.pattern = cv2.medianBlur(self.pattern,11)
        self.pattern = cv2.blur(self.pattern,(11,11))
        
        ## @var sample.
        # User's processed images samples.
        self.samples = []
        
        ## @var sample.
        # User's samples names.
        self.sample_name = []
        
        ## @var sift.
        # SIFT opencv object.
        self.sift = cv2.xfeatures2d.SIFT_create(number_of_keypoints)
        
        ## @var des_P.
        # Keypoints and descriptors of the pattern.
        self.kp_P, self.des_P = self.sift.detectAndCompute(self.pattern,None)
        
        ## @var des_samples.
        # User's keypoints and descriptors of the samples.
        self.des_samples = []
        
        ## @var good_matches_Mated.
        # User's array of good matches of the Mated comparation.
        self.good_matches_Mated = []
        
        ## @var good_matches_NonMated.
        # User's array of good matches of the Non-Mated comparation.
        self.good_matches_NonMated = []
        
    ## Function to define the keypoints of all the users' samples.
    #  @return Void.    
    def define_key_points(self):
        for i in self.samples:
            kp, des = self.sift.detectAndCompute(i,None)
            self.des_samples.append(des)        

## Class to execute the SIFT algorithm and obtain the score of the matching, Mated and Non-Mated, in the '.txt' files.  
#           
class SIFT:
    ## Constructor
    #  @arg self The object pointer.
    #  @return Void.
    def __init__(self):
        
        ## @var matches.
        # Array of all matches of the Mated comparation.
        self.matches_Mated = []
        
        ## @var matches.
        # Array of all matches of the Non-Mated comparation.
        self.matches_NonMated = []
        
        ## @var good_matches_Mated.
        # Array of good matches of the Mated comparation.
        self.good_matches_Mated = []
        
        ## @var good_matches_NonMated.
        # Array of good matches of the Non-Mated comparation.
        self.good_matches_NonMated = []
        
        ## @var started.
        # Flag to indicate that the thread has been started.
        self.started = False
    
    ## Function to start the thread to execute the program in background.
    #  @arg users Array of user object.
    #  @return The object pointer.
    def start(self, users):
        if self.started:
            print ("Already started!!")
            return None
        self.started = True
        
        ## @var users.
        # Array of user object.
        self.users = users
        
        ## @var thread.
        # Thread object.
        self.thread = Thread(target=self.run, args=())
        
        # Thread starts.
        self.thread.start()
        return self
    
    ## Function to run the SIFT algorithm and the matching.
    #  @return Void.
    def run(self):
            
            ## @var FLANN_INDEX_KDTREE.
            # Index parameter for SIFT, SURF, etc., algorithms to use with Flann Based Macher. See OpenCV documentation.
            self.FLANN_INDEX_KDTREE = 0
            
            ## @var index_params.
            # Index parameters for SIFT, SURF, etc., algorithms to use with Flann Based Macher. See OpenCV documentation.
            self.index_params = dict(algorithm = self.FLANN_INDEX_KDTREE, trees = 5)
            
            ## @var search_params.
            # Search parameters for SIFT, SURF, etc., algorithms to use with Flann Based Macher. See OpenCV documentation..
            self.search_params = dict(checks = 50)
            
            ## @var flann.
            # Flann Based Macher initialization (method to compare the matches obtained).
            self.flann = cv2.FlannBasedMatcher(self.index_params, self.search_params)
            
            # Run total timer to know the time to obtain the matching of the entire database (Mated and Non-Mated comparation).
            start_total = time.time()
            
            # Run total timer to know the time to obtain the matching of Mated comparation.
            start = time.time()
            
            # Mated comparation.
            
            # Evaluate all users.
            for x in range (self.users.number_of_users):
                self.matches_Mated.clear()                                                                        # Reset Mated matches array.
                # Evaluate all samples of the user.
                for i in range(len(self.users.users[x].des_samples)):
                    # Matching between the pattern descriptors and the sample descriptors.
                    self.matches_Mated.insert(i, self.flann.knnMatch(self.users.users[x].des_P, self.users.users[x].des_samples[i],k=2))
                    self.good_matches_Mated.clear()                                                               # Reset good Mated matches array.
                    
                    # Store all the good matches as per Lowe's ratio test.
                    for m,n in self.matches_Mated[i]:
                        if m.distance < Lowes_ratio_test * n.distance:
                            self.good_matches_Mated.append(m)                                                      # Save good Mated matches.
                            
                    self.users.users[x].good_matches_Mated.append(len(self.good_matches_Mated))                    # Save good Mated matches in the object.
            
            # Stop end timer to know the time to obtain the matching of Mated comparation.
            end = time.time()
            print("MATED TIME(s):", end - start)
            
            # Run total timer to know the time to obtain the matching of Non-Mated comparation.
            start = time.time()
            
            # Non-Mated comparation.
            
            # Initialize good Non-Mated array of the object.
            for x in range (self.users.number_of_users):
                # Evaluate all the others users.
                for i in range(self.users.number_of_users):
                    self.users.users[x].good_matches_NonMated.append([])                                           # Reset Non-Mated matches array.
                    for j in range(len(self.users.users[i].des_samples)):
                        if(x != i):
                         self.users.users[x].good_matches_NonMated[i].append(j)
            
            # Evaluate all users.
            for x in range (self.users.number_of_users):
                self.matches_NonMated.clear()                                                                      # Reset Non-Mated matches array.
                # Evaluate all the others users.
                for i in range(self.users.number_of_users):
                    # Evaluate all the others users and its sammples.
                    for j in range(len(self.users.users[i].des_samples)):
                        if(x != i):
                            # Matching between the pattern descriptors and the others users patterns and samples descriptors.
                            self.matches_NonMated.insert(j, self.flann.knnMatch(self.users.users[x].des_P, self.users.users[i].des_samples[j],k=2))
                            self.good_matches_NonMated.clear()                                                     # Reset good Non-Mated matches array.
                            
                            # Store all the good matches as per Lowe's ratio test.
                            for m,n in self.matches_NonMated[j]:
                                if m.distance < Lowes_ratio_test * n.distance:
                                    self.good_matches_NonMated.append(m)                                           # Save good Non-Mated matches.
                                
                            self.users.users[x].good_matches_NonMated[i].insert(j,len(self.good_matches_NonMated)) # Save good Non-Mated matches in the object.
            
            # Stop end timer to know the time to obtain the matching of Non-Mated comparation.
            end = time.time()
            print("NON-MATED TIME(s):", end - start)
            
            # Stop total timer to know the time to obtain the matching of the entire database (Mated and Non-Mated comparation).
            end_total = time.time()
            print("TOTAL TIME(s):", end_total - start_total)
            
            # Write the results in the '.txt' files.
            self.write_mated_file()
            self.write_non_mated_file()
    
    ## Function to create, open and write the Mated '.txt' file. 
    #  @return Void.
    def write_mated_file(self):
        # Open file.
        self.open_file = open(mated_txt_name, "w", newline="")
        self.writer = csv.writer(self.open_file, delimiter=";")
        
        # Write file.
        for x in range (self.users.number_of_users):
            for i in range(len(self.users.users[x].samples)):
                self.writer.writerow(["", self.users.users[x].name, self.users.users[x].sample_name[i], self.users.users[x].good_matches_Mated[i]])
        
        # Close file.
        self.open_file.close()

    ## Function to create, open and write the Non-Mated '.txt' file. 
    #  @return Void.
    def write_non_mated_file(self):
        # Open file.
        self.open_file = open(non_mated_txt_name, "w", newline="")
        self.writer = csv.writer(self.open_file, delimiter=";")
        
        # Write file.
        for x in range (self.users.number_of_users):
            for i in range (self.users.number_of_users):
                for j in range(len(self.users.users[i].des_samples)):
                    if(x != i):
                        self.writer.writerow(["", self.users.users[x].name, str(self.users.users[i].name) + "-" + str(self.users.users[i].sample_name[j]), self.users.users[x].good_matches_NonMated[i][j]])
                        
        # Close file.
        self.open_file.close()
        
    ## Function to stop the thread.
    #  @return Void.
    def stop(self) :
        self.started = False
        self.thread.join()


# EXECUTION

# Users definition.
users_object = DefineUsers()
# Define users' attributes.
users_object.define_users()

# SIFT algorithm and matching.
sift_algorithm = SIFT().start(users_object.get_users())

